@@ -26,6 +26,7 @@
 #include <string.h>
 #include "gok-key.h"
 #include "gok-keyboard.h"
+#include "gok-mousecontrol.h"
 #include "gok-log.h"
 #include "gok-modifier.h"
 #include "gok-scanner.h"
@@ -33,6 +34,8 @@
 #include "gok-data.h"
 #include "gok-branchback-stack.h"
 #include "gok-modifier-keymasks.h"
+#include "gok-gconf-keys.h"
+#include "main.h"
 
 #define XML_STRING_SIZE_STOCK ((const xmlChar *) "stock")
 #define XML_STRING_SIZE_FIXED ((const xmlChar *) "fixed")
@@ -123,7 +126,7 @@ gok_key_lookup_shorter_label (char *str)
 	return str;
 }
 
-static char *
+static char *	
 gok_key_label_from_keysym_string (const char *str)
 {
 	char *cp;
@@ -328,7 +331,7 @@ gok_key_set_effective_group (gint group)
 }
 
 void
-gok_keyimage_set_size_from_spec (GokKeyImage *keyimage, gchar *sizespec)
+gok_keyimage_set_size_from_spec (GokKeyImage *keyimage, gchar *sizespec, gchar *align)
 {
 	if (xmlStrPrefix (sizespec, XML_STRING_SIZE_STOCK)) {
 		keyimage->type = IMAGE_TYPE_STOCK;
@@ -340,19 +343,24 @@ gok_keyimage_set_size_from_spec (GokKeyImage *keyimage, gchar *sizespec)
 			keyimage->w = w;
 			keyimage->h = h;
 		}
-		fprintf (stderr, "image size: %d,%d\n", keyimage->w, keyimage->h);
 	}
 	else if (xmlStrPrefix (sizespec, XML_STRING_SIZE_FIT))
 		keyimage->type = IMAGE_TYPE_FIT;
 	else if (xmlStrPrefix (sizespec, XML_STRING_SIZE_FILL))
 		keyimage->type = IMAGE_TYPE_FILL;
 
+	if (!xmlStrcmp (align, (const xmlChar *) "right")) {
+		keyimage->placement_policy = IMAGE_PLACEMENT_RIGHT;
+	}
+	else { /* TODO: that's all we recognize for now, implement other options */
+		keyimage->placement_policy = IMAGE_PLACEMENT_LEFT;
+	}
 }
 
 GokKey *
 gok_key_from_xkb_key (GokKey *prevKey, GokKeyboard *pKeyboard, Display *display, XkbGeometryPtr pGeom, XkbRowPtr pRow, XkbSectionPtr pXkbSection, XkbKeyPtr keyp, int section, int row, int col)
 {
-  GokKey *pKey = gok_key_new (prevKey, NULL, pKeyboard);
+  GokKey *pKey;
   char *label;
   char name[XkbKeyNameLength + 1];
   char keycode_name[8];
@@ -362,6 +370,9 @@ gok_key_from_xkb_key (GokKey *prevKey, GokKeyboard *pKeyboard, Display *display,
   gboolean is_modifier = FALSE;
   unsigned int numlock_mask = gok_key_get_numlock_mask (display);
 
+  pKey = gok_key_new (prevKey, NULL, pKeyboard);
+  numlock_mask = gok_key_get_numlock_mask (display);
+  is_modifier = FALSE;
   pKey->has_text = TRUE; /* TODO: handle non-textual keyboard keys! */
   pKey->Top = row;
   pBounds = &pGeom->shapes[keyp->shape_ndx].bounds;
@@ -377,10 +388,15 @@ gok_key_from_xkb_key (GokKey *prevKey, GokKeyboard *pKeyboard, Display *display,
 	  is_modifier = TRUE;
 	  gok_modifier_add (pKey->ModifierName);
   }
-  if (is_modifier) /* FIXME */
+  if (is_modifier) { /* FIXME */
     pKey->Type = KEYTYPE_MODIFIER;
-  else
+    pKey->pImage = gok_keyimage_new (pKey, NULL);
+    pKey->pImage->type = IMAGE_TYPE_INDICATOR;
+  }
+  else {
     pKey->Type = KEYTYPE_NORMAL;
+	  pKey->is_repeatable = TRUE;
+  }
   pKey->Style = KEYSTYLE_NORMAL;
   label = gok_key_label_from_keycode (keycode, display, 0, 0);
   pKey->FontSize = -1;
@@ -496,9 +512,13 @@ GokKey* gok_key_new (GokKey* pKeyPrevious, GokKey* pKeyNext, GokKeyboard* pKeybo
 	pgok_key_new->FontSizeGroup = FONT_SIZE_GROUP_UNDEFINED;
 	pgok_key_new->FontSize = -1; /* not set yet */
 	pgok_key_new->ComponentState.active = 0;
+	pgok_key_new->ComponentState.radio = 0;
+	pgok_key_new->ComponentState.latched = 0;
+	pgok_key_new->ComponentState.locked = 0;
 	pgok_key_new->State = GTK_STATE_NORMAL;
 	pgok_key_new->StateWhenNotFlashed = GTK_STATE_NORMAL;
 	pgok_key_new->pGeneral = NULL;
+	pgok_key_new->is_repeatable = FALSE;
 	
 	return pgok_key_new;
 }
@@ -607,23 +627,30 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 	xmlNode* pNodeWrapperChild;
 	GokOutput* pNewOutput;
 	GokOutput* pOutputTemp;
-
+	gboolean settings_locked;
+	
 	g_assert (pKey != NULL);
 	g_assert (pNode != NULL);
 
 	pKey->has_text = FALSE;
+	pKey->is_repeatable = FALSE;
 
+	/* TODO: make more styles in gok.rc */
+	
 	/* type of key (normal, branch etc.) */
 	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "type");
 	if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"normal") == 0)
 	{
 		pKey->Type = KEYTYPE_NORMAL;
 		pKey->Style = KEYSTYLE_NORMAL;
+		pKey->is_repeatable = TRUE;
 	}
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *) "modifier") == 0)
 	{
 		pKey->Type = KEYTYPE_MODIFIER;
 		pKey->Style = KEYSTYLE_NORMAL;
+		pKey->pImage = gok_keyimage_new (pKey, NULL);
+		pKey->pImage->type = IMAGE_TYPE_INDICATOR;
 	}
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *) "branch") == 0)
 	{
@@ -662,8 +689,8 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 	}
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchEditText") == 0)
 	{
-		pKey->Type = KEYTYPE_BRANCHEDITTEXT;
-		pKey->Style = KEYSTYLE_BRANCHEDITTEXT;
+		pKey->Type = KEYTYPE_BRANCHCOMPOSE;
+		pKey->Style = KEYSTYLE_GENERALDYNAMIC;
 	}
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchSettings") == 0)
 	{
@@ -673,14 +700,25 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"settings") == 0)
 	{
 		pKey->Type = KEYTYPE_SETTINGS;
-		pKey->Style = KEYSTYLE_SETTINGS;
+		if (gok_main_get_login ())
+		{
+				pKey->Style = KEYSTYLE_INSENSITIVE;
+		}
+		else {
+			gok_gconf_get_bool (gconf_client_get_default(),
+				GOK_GCONF_PREFS_LOCKED, &settings_locked);
+			if (settings_locked) {
+				gok_settingsdialog_lock ();
+			}
+			pKey->Style = KEYSTYLE_SETTINGS;
+		}
 	}
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchWindows") == 0)
 	{
 		pKey->Type = KEYTYPE_BRANCHWINDOWS;
 		pKey->Style = KEYSTYLE_BRANCH;
 	}
-	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"mouse") == 0)
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"pointer") == 0)
 	{
 		pKey->Type = KEYTYPE_POINTERCONTROL;
 		pKey->Style = KEYSTYLE_POINTERCONTROL;
@@ -712,6 +750,7 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 				gok_keyboard_parse_direction (pStringAttributeValue);
 			g_free (pStringAttributeValue);
 		}
+		pKey->is_repeatable = TRUE;
 	}
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"help") == 0)
 	{
@@ -723,6 +762,56 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 		pKey->Type = KEYTYPE_ABOUT;
 		pKey->Style = KEYSTYLE_ABOUT;
 	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"navigate") == 0)
+	{
+		pKey->Type = KEYTYPE_TEXTNAV;
+		pKey->Style = KEYSTYLE_NORMAL;
+		pKey->is_repeatable = TRUE;
+		gok_compose_key_init (pKey, pNode);
+	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"edit") == 0)
+	{
+		pKey->Type = KEYTYPE_EDIT;
+		pKey->Style = KEYSTYLE_NORMAL;
+		gok_compose_key_init (pKey, pNode);
+	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"select") == 0)
+	{
+		pKey->Type = KEYTYPE_SELECT;
+		pKey->Style = KEYSTYLE_SELECT;
+		gok_compose_key_init (pKey, pNode);
+	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"select-toggle") == 0)
+	{
+		pKey->Type = KEYTYPE_TOGGLESELECT;
+		pKey->Style = KEYSTYLE_SELECT;
+		pKey->pImage = gok_keyimage_new (pKey, NULL);
+		pKey->pImage->type = IMAGE_TYPE_INDICATOR;
+		pKey->has_image = TRUE;
+		pKey->ComponentState.active = FALSE;
+		gok_compose_key_init (pKey, pNode);
+	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"repeat-next") == 0)
+	{
+		pKey->Type = KEYTYPE_REPEATNEXT;
+		pKey->Style = KEYSTYLE_NORMAL;
+	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"mouse") == 0)
+	{
+		pKey->Type = KEYTYPE_MOUSE;
+		pKey->Style = KEYSTYLE_NORMAL;
+		pKey->is_repeatable = TRUE;
+		gok_mouse_control_init (pKey, pNode);
+	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"mousebutton") == 0)
+	{
+		pKey->Type = KEYTYPE_MOUSEBUTTON;
+		if (!gok_scanner_current_state_uses_corepointer ()) 
+			pKey->Style = KEYSTYLE_MOUSEBUTTON;
+		else
+			pKey->Style = KEYSTYLE_INSENSITIVE;
+		gok_mouse_control_init (pKey, pNode);
+	}
 
 	/* branch target */
 	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "target");
@@ -829,9 +918,10 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 			/* TODO: support href as well as local files ? */
 			gchar *filename = xmlGetProp (pNodeKeyChild, (const xmlChar *) "source");
 			gchar *sizespec = xmlGetProp (pNodeKeyChild, (const xmlChar *) "type");
+			gchar *align = xmlGetProp (pNodeKeyChild, (const xmlChar *) "align");
 			gok_keyimage_new (pKey, filename);
 			if (sizespec != NULL) 
-				gok_keyimage_set_size_from_spec (pKey->pImage, sizespec);
+				gok_keyimage_set_size_from_spec (pKey->pImage, sizespec, align);
 			pKey->has_image = TRUE;
 		}
 
@@ -1175,6 +1265,39 @@ gok_key_get_image (GokKey* pKey)
 	return pKey->pImage;
 }
 
+GtkWidget *
+gok_key_status_image (GokKey *key) 
+{
+	GokUIState state = key->ComponentState;
+	GtkWidget *image = NULL;
+	gchar *icon_name, *file;
+
+#if STATUS_IMAGES
+	if (state.latched) {
+		icon_name = "gok/latched.png";
+	}
+	else if (state.locked) {
+		icon_name = "gok/locked.png";
+	}
+	else {
+		icon_name = "gok/empty.png";
+	}
+#else
+	icon_name = "gok/small-empty.png";
+#endif
+	file = gnome_program_locate_file (NULL, GNOME_FILE_DOMAIN_APP_DATADIR,
+					  icon_name,
+					  FALSE, NULL);
+
+	if (file) {
+		image = gtk_image_new_from_file (file);
+		g_free (file);
+	}
+
+	return image;
+}
+	
+
 /**
 * gok_key_create_image_widget
 * @pKey: Pointer to the key that you want the image for.
@@ -1185,17 +1308,19 @@ GtkWidget*
 gok_key_create_image_widget (GokKey* pKey)
 {
 	GtkWidget *image = NULL;
+	gchar *file;
 	if (pKey->pImage != NULL) {
 		/* FIT and FILL aren't yet implemented, since they are dynamic */
 		switch (pKey->pImage->type) {
+		case IMAGE_TYPE_INDICATOR:
+			image = gok_key_status_image (pKey);
+			break;
 		case IMAGE_TYPE_STOCK:
 			image = gtk_image_new_from_stock (pKey->pImage->Filename,
 							  pKey->pImage->stock_size);
 			break;
 		case IMAGE_TYPE_FIXED:
 		default:
-			fprintf (stderr, "scaled image %d,%d\n", 
-				 pKey->pImage->w, pKey->pImage->h);
 			if ((pKey->pImage->w == -1) && (pKey->pImage->h == -1)) {
 				image = gtk_image_new_from_file (
 					gok_key_get_image_filename (pKey));
@@ -1388,7 +1513,6 @@ void gok_key_set_button_name (GokKey* pKey)
 	switch (pKey->Style)
 	{
 	case KEYSTYLE_DOCK:
-	case KEYSTYLE_POINTERCONTROL:
 	case KEYSTYLE_NORMAL:
 		strcpy (styleButton, "StyleButtonNormal");
 		strcpy (styleText, "StyleTextNormal");
@@ -1415,11 +1539,16 @@ void gok_key_set_button_name (GokKey* pKey)
 			strcpy (styleText, "StyleTextNormal");
 		}
 		break;
-		
+
 	case KEYSTYLE_GENERALDYNAMIC:
 		strcpy (styleButton, "StyleButtonGeneralDynamic");
 		strcpy (styleText, "StyleTextNormal");
 		break;
+
+	case KEYSTYLE_SELECT:
+		strcpy (styleButton, "StyleButtonSelect");
+		strcpy (styleText, "StyleTextNormal");
+		break;
 		
 	case KEYSTYLE_BRANCHMENUS:
 		strcpy (styleButton, "StyleButtonBranchMenus");
@@ -1448,8 +1577,9 @@ void gok_key_set_button_name (GokKey* pKey)
 		break;
 
 	case KEYSTYLE_BRANCHGUIACTIONS:
+	case KEYSTYLE_PAGESELECTION:
 	case KEYSTYLE_HYPERLINK:
-	case KEYSTYLE_EDITABLETEXTACTION:
+	case KEYSTYLE_EDIT:
 		strcpy (styleButton, "StyleButtonBranchGuiActions");
 		strcpy (styleText, "StyleTextNormal");
 		break;
@@ -1470,7 +1600,9 @@ void gok_key_set_button_name (GokKey* pKey)
 		strcpy (styleText, "StyleTextHelp");
 		break;
 					
+	case KEYSTYLE_POINTERCONTROL:
 	case KEYSTYLE_WORDCOMPLETE:
+	case KEYSTYLE_SPELL:
 		strcpy (styleButton, "StyleButtonWordComplete");
 		strcpy (styleText, "StyleTextWordComplete");
 		break;
@@ -1481,8 +1613,8 @@ void gok_key_set_button_name (GokKey* pKey)
 		break;
 	default:
 		gok_log_x ("Warning: default hit in gok_key_set_button_name! Key = %s\n", gok_key_get_label (pKey));
-		strcpy (styleButton, "StyleButtonDisabled");
-		strcpy (styleText, "StyleTextDisabled");
+		strcpy (styleButton, "StyleButtonNormal");
+		strcpy (styleText, "StyleTextNormal");
 		break;
 	}
 	
@@ -1497,7 +1629,7 @@ void gok_key_set_button_name (GokKey* pKey)
 
 /**
  * gok_key_update_toggle_state:
- *
+ * @pKey: pointer to the key to update.
  *
  **/
 void
@@ -1510,6 +1642,8 @@ gok_key_update_toggle_state (GokKey *pKey)
 			(gok_modifier_get_state (pKey->ModifierName) != MODIFIER_STATE_OFF);
 		else 
 			is_active = pKey->ComponentState.active;
+		if (pKey->ComponentState.radio) 
+			GOK_BUTTON (pKey->pButton)->indicator_type = "radiobutton";
 		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (pKey->pButton),
 					      is_active);
 	}
@@ -2001,7 +2135,7 @@ gok_keyimage_new (GokKey* pKey, gchar* pFilename)
 	}
 	
 	return pNewImage;
-	}
+}
 	 
 /**
 * gok_keyimage_delete
@@ -2029,3 +2163,26 @@ gok_key_get_image_filename (GokKey* pKey)
 	return pKey->pImage->Filename;
 }
 
+/**
+* gok_key_isRepeatable
+* @pKey: Pointer to the key to assess..
+*
+* Call this to see if the key is repeatable.
+*
+* returns: gboolean
+**/
+gboolean 
+gok_key_isRepeatable(GokKey* pKey)
+{
+	gboolean returncode = FALSE;
+	gok_log_enter();
+	if (pKey != NULL)
+	{
+		if ( pKey->is_repeatable ) {
+			returncode = TRUE;
+			gok_log("this key is repeatable");
+		}		
+	}
+	gok_log_leave();
+	return returncode;
+}
