@@ -364,7 +364,8 @@ gok_keyboard_add_keys_from_xkb_geom (GokKeyboard *pKeyboard, XkbDescPtr kbd)
 	/* add a branch key for special text edit functions */
 	pKey = gok_key_new (pKey, NULL, pKeyboard);
 	pKey->Style = KEYSTYLE_GENERALDYNAMIC;
-	pKey->Type = KEYTYPE_BRANCHEDITTEXT;
+	pKey->Type = KEYTYPE_BRANCHCOMPOSE;
+	pKey->Target = "text-operations";
 	pKey->Top = row + 1;
 	pKey->Bottom = row + 2;
 	pKey->Left = 0;
@@ -1274,10 +1275,10 @@ gboolean gok_keyboard_display (GokKeyboard* pKeyboard, GokKeyboard* pKeyboardCur
 	strcpy (titleWindow, _("GOK - "));
 	strcat (titleWindow, pKeyboard->Name);
 	gtk_window_set_title (GTK_WINDOW(pWindowMain), titleWindow);
-
+	
 	/* get the "fixed container" that holds the buttons */
 	pFixedContainer = GTK_BIN(pWindowMain)->child;
-
+	
 	/* create all the buttons and add them to the container */
 	pKey = pKeyboard->pKeyFirst;
 	while (pKey != NULL)
@@ -1288,15 +1289,22 @@ gboolean gok_keyboard_display (GokKeyboard* pKeyboard, GokKeyboard* pKeyboardCur
 			/* create a new GOK button */
 			if (pKey->has_text)
 			{
-				pNewButton = gok_button_new_with_label (gok_key_get_label (pKey)); 
+				pNewButton = gok_button_new_with_label (gok_key_get_label (pKey), IMAGE_PLACEMENT_LEFT); 
 				if (pKey->has_image) {
 					gok_button_set_image (GOK_BUTTON (pNewButton), 
 							      GTK_IMAGE (gok_key_create_image_widget (pKey)));
+					if (pKey->pImage->type != IMAGE_TYPE_INDICATOR) {
+						GOK_BUTTON (pNewButton)->indicator_type = NULL;
+					}
 				}
 			}
 			else if (pKey->has_image)
 			{
-				pNewButton = gok_button_new_with_image (gok_key_create_image_widget (pKey));
+				pNewButton = gok_button_new_with_image (gok_key_create_image_widget (pKey), 
+									pKey->pImage->placement_policy);
+				if (pKey->pImage->type != IMAGE_TYPE_INDICATOR) {
+					GOK_BUTTON (pNewButton)->indicator_type = NULL;
+				}
 			}
 			else {
 			        pNewButton = g_object_new (GOK_TYPE_BUTTON, NULL);
@@ -1533,6 +1541,26 @@ void gok_keyboard_position_keys (GokKeyboard* pKeyboard, GtkWidget* pWindow)
 	}
 }
 
+static gboolean
+gok_keyboard_page_select (AccessibleNode *node) 
+{
+	AccessibleSelection *selection;
+	Accessible *parent;
+	int index;
+	gboolean retval = FALSE;
+       
+	parent = Accessible_getParent (node->paccessible);
+	if (parent) {
+		index = Accessible_getIndexInParent (node->paccessible);
+		selection = Accessible_getSelection (parent);
+		g_assert (selection != NULL);
+		retval = AccessibleSelection_selectChild (selection, index);
+	}
+	
+	return retval;
+}
+
+
 /**
 * gok_keyboard_branch_byKey
 * @pKey: The key that is causes the branch.
@@ -1583,6 +1611,11 @@ gboolean gok_keyboard_branch_byKey (GokKey* pKey)
 						       GOK_SPY_SEARCH_UI,
 						       SPI_ROLE_PUSH_BUTTON); 
 			break;			
+		case KEYTYPE_PAGESELECTION:
+			gok_log("page select");
+			gok_keyboard_page_select (pKey->pNodeAccessible);
+			return FALSE;
+			break;
 		case KEYTYPE_BRANCHGUIACTIONS:
 			gok_log("branch gui_actions");
 			is_branched = gok_keyboard_branch_gui_actions(pKey->pNodeAccessible);
@@ -1596,10 +1629,6 @@ gboolean gok_keyboard_branch_byKey (GokKey* pKey)
 			}
 			return is_branched;
 			break;
-		case KEYTYPE_BRANCHEDITTEXT:
-			gok_log("branch edittext");		
-			return gok_keyboard_branch_edittext(); 
-			break;
 	        case KEYTYPE_BRANCHHYPERTEXT:
 			fprintf (stderr, "HYPERTEXT!\n");
                         break;
@@ -1767,12 +1796,17 @@ gboolean gok_keyboard_update_dynamic (GokKeyboard* pKeyboard)
 			pKey->ComponentState.active = is_active;
 			pKey->ComponentState.radio = TRUE;
 			break;
+		case SPI_ROLE_PAGE_TAB:
+			/* no action implemented,, must use selection API on parent */
+			pKey->Type = KEYTYPE_PAGESELECTION;
+			pKey->Style = gok_style_if_enabled (pStateSet, KEYSTYLE_PAGESELECTION);
+			break;
 		case SPI_ROLE_TEXT:
 			/* should only be in the list if it's editable...*/
-			if (pNodeAccessible->is_link) {
+			if (pNodeAccessibleNext->is_link) {
 			      pKey->Type = KEYTYPE_HYPERLINK;
 			      pKey->Style = KEYSTYLE_HYPERLINK;
-			      is_active = TRUE;
+			      is_active = FALSE;
 			}
 			else {
 			      pKey->Type = KEYTYPE_BRANCHTEXT;
@@ -1943,6 +1977,7 @@ gboolean gok_keyboard_branch_gui (AccessibleNode* pNodeAccessible,
 	return TRUE;
 }
 
+
 /**
 * gok_keyboard_branch_gui_actions
 * @pNodeAccessible: the node thich represents the gui widget
@@ -1967,7 +2002,8 @@ gboolean gok_keyboard_branch_gui_actions (AccessibleNode* pNodeAccessible)
 	/* 
 	 * just in case: before we do any actions check to see if this object has 
 	 * interesting children and if so, branch to them instead.
-	 *
+	 * Note that this may not be the right thing to do in all cases,
+	 * since it implies that we can't activate the parent's actions.
 	 */
 	gok_log("checking for children with interesting roles");
 	for (i=0; i< NUM_INTERESTING_ROLES; i++)
@@ -2007,18 +2043,13 @@ gboolean gok_keyboard_branch_gui_actions (AccessibleNode* pNodeAccessible)
 	return branch;	
 }
 
-/**
-* gok_keyboard_branch_edittext
-*
-* returns: TRUE if the keyboard was displayed, FALSE if not.
-*/
-gboolean gok_keyboard_branch_edittext ()
-{
-	return gok_composer_show();
-}
-
 gboolean gok_keyboard_branch_editableTextAction (GokKeyboard* pKeyboard, GokKey* pKey)
 {
+	/* FIXME: this is suspect */
+	pKeyboard->pAccessible = gok_spy_get_accessibleWithText ();
+	if (pKeyboard->pAccessible) {
+		Accessible_ref (pKeyboard->pAccessible);
+	}
 	return gok_composer_branch_textAction (pKeyboard, pKey);
 }
 	
@@ -2360,17 +2391,41 @@ gok_keyboard_follow_link (GokKey *key)
 		/* can't activate editable text this way, it will insert a space instead! */
 		if (hypertext && text && node->is_link && !is_editable) {
 			AccessibleHyperlink *link = AccessibleHypertext_getLink (hypertext, node->link);
+			Accessible *target;
+			AccessibleAction *action;
+			gboolean is_action = FALSE;
+			gboolean is_focussed = FALSE;
 			long int start, end;
-			gboolean focussed;
 			AccessibleHyperlink_getIndexRange (link, &start, &end);
-			focussed = gok_keyboard_focus_object (node->paccessible);
-			if (focussed) {
+			is_focussed = gok_keyboard_focus_object (node->paccessible);
+			if (!is_focussed) gok_log_x ("Could not focus object\n");
+			else {
 				AccessibleText_setCaretOffset (text, start);
 				SPI_generateKeyboardEvent ((long) XK_space, NULL, SPI_KEY_SYM);
 			}
+			target = AccessibleHyperlink_getObject (link, 0);
+			if (target) {
+				action = Accessible_getAction (target);
+				if (!action) { /* try the parent object? */
+					action = Accessible_getAction (link);
+					if (action) gok_log_x ("activating the link object.\n");
+				}
+				if (!action) { /* try the parent object? */
+					action = Accessible_getAction (node->paccessible);
+					if (action) gok_log_x ("activating the parent hypertext object!\n");
+				}
+				if (action) {
+					AccessibleAction_doAction (action, 0); /* HACK, should check the action first */
+					gok_log ("activating link\n");
+					AccessibleAction_unref (action);
+					is_action = TRUE;
+				}
+				Accessible_unref (target);
+			}
 			AccessibleText_unref (text);
+			AccessibleHyperlink_unref (link);
 			AccessibleHypertext_unref (text);
-			return focussed;
+			return is_action;
 		}
 	}
 	return FALSE;
@@ -2520,8 +2575,8 @@ gok_keyboard_dock (GokKey *pKey)
  *        should relocate the GokKeyboard's default position 
  *        onscreen.
  *
- * Moves the GOK keyboard some distance according to the 
- * value of the 'direction' data in 'pKey->pGeneral'.
+ * Moves the GOK keyboard some distance, or resizes its keys, 
+ * according to the value of the 'direction' data in 'pKey->pGeneral'.
  *
  **/
 void
@@ -2529,6 +2584,7 @@ gok_keyboard_move_resize (GokKey *pKey)
 {
 	GokKeyboardDirection dir = GOK_DIRECTION_NONE;
 	GtkWidget *window = gok_main_get_main_window ();
+	gboolean resize = FALSE;
 	gint x, y;
 
 	gok_log_enter();
@@ -2560,14 +2616,40 @@ gok_keyboard_move_resize (GokKey *pKey)
 		break;
 	case GOK_DIRECTION_FILL_EW:
 		/* toggle expand */
-		fprintf (stderr, "toggling fill\n");
 		gok_data_set_expand (!gok_data_get_expand ());
-		gok_main_display_scan_reset ();
+		resize = TRUE;
+		break;
+	case GOK_RESIZE_NARROWER:
+		x = gok_data_get_key_width ();
+		x -= 2;
+		gok_data_set_key_width (MAX (0, x));
+		resize = TRUE;
+		break;
+	case GOK_RESIZE_WIDER:
+		gok_data_set_key_width (gok_data_get_key_width () + 2); 
+		resize = TRUE;
+		break;
+	case GOK_RESIZE_SHORTER:
+		y = gok_data_get_key_height ();
+		y -= 2;
+		gok_data_set_key_height (MAX (0, y));
+		resize = TRUE;
+		break;
+	case GOK_RESIZE_TALLER:
+		gok_data_set_key_height (gok_data_get_key_height () + 2);
+		resize = TRUE;
 		break;
 	default:
 	        break;
 	}
-	gtk_window_move (GTK_WINDOW (window), x, y);
+	if (resize) {
+		gok_keyboard_display (gok_main_get_current_keyboard (),
+				      gok_main_get_current_keyboard (), 
+				      gok_main_get_main_window (), TRUE);
+	}
+	else {
+		gtk_window_move (GTK_WINDOW (window), x, y);
+	}
 	gok_log_leave();
 }
 
@@ -2626,6 +2708,14 @@ GokKey* gok_keyboard_output_key(GokKey* pKey)
 
 	switch (pKey->Type)
 	{
+		case KEYTYPE_BRANCHCOMPOSE:
+			if (! gok_spy_get_accessibleWithText ()) {
+				gok_log_x ("object isn't a text object, can't branch to compose");
+				gok_log_leave ();
+				return NULL;
+			}
+			gok_composer_validate (pKey->Target, gok_spy_get_accessibleWithText ());
+			/* else fall-through */
 		case KEYTYPE_BRANCH:
 			gok_main_display_scan ( NULL, pKey->Target, 
 				KEYBOARD_TYPE_UNSPECIFIED, KEYBOARD_LAYOUT_NORMAL);
@@ -2663,7 +2753,6 @@ gok_log_x("BRANCH BACK %s", pKey->pLabel->Text);
 		case KEYTYPE_BRANCHMENUS:
 		case KEYTYPE_BRANCHTOOLBARS:
 		case KEYTYPE_BRANCHGUI:
-		case KEYTYPE_BRANCHEDITTEXT:
 		case KEYTYPE_BRANCHHYPERTEXT:
 			gok_keyboard_branch_byKey (pKey);
 			break;
@@ -2738,6 +2827,7 @@ gok_log_x("BRANCH BACK %s", pKey->pLabel->Text);
 			
 			/* turn off any modifier keys that are not locked on */
 			gok_modifier_all_off();
+			gok_wordcomplete_increment_word_frequency (gok_key_get_label (pKey));
 			
 			/* reset the word completor */
 			gok_wordcomplete_clear_keys();
@@ -2746,28 +2836,16 @@ gok_log_x("BRANCH BACK %s", pKey->pLabel->Text);
 			
 		case KEYTYPE_BRANCHMENUITEMS:
 		case KEYTYPE_MENUITEM:
+		case KEYTYPE_PAGESELECTION:
 		case KEYTYPE_BRANCHGUIACTIONS:
 			if (!gok_keyboard_branch_byKey (pKey))
 				gok_main_display_scan_reset ();
 			break;
 			
-		case KEYTYPE_ACCESSIBLETEXT_CHARRIGHT:
-		case KEYTYPE_ACCESSIBLETEXT_CHARLEFT:
-		case KEYTYPE_ACCESSIBLETEXT_WORDRIGHT:
-		case KEYTYPE_ACCESSIBLETEXT_WORDLEFT:
-		case KEYTYPE_ACCESSIBLETEXT_SENTENCERIGHT:
-		case KEYTYPE_ACCESSIBLETEXT_SENTENCELEFT:
-		case KEYTYPE_ACCESSIBLETEXT_SELECT_CHARRIGHT:
-		case KEYTYPE_ACCESSIBLETEXT_SELECT_CHARLEFT:
-		case KEYTYPE_ACCESSIBLETEXT_SELECT_WORDRIGHT:
-		case KEYTYPE_ACCESSIBLETEXT_SELECT_WORDLEFT:
-		case KEYTYPE_ACCESSIBLETEXT_SELECT_SENTENCERIGHT:
-		case KEYTYPE_ACCESSIBLETEXT_SELECT_SENTENCELEFT:
-		case KEYTYPE_EDITABLETEXT_COPY:
-		case KEYTYPE_EDITABLETEXT_DELETE:
-		case KEYTYPE_EDITABLETEXT_CUT:
-		case KEYTYPE_EDITABLETEXT_PASTE:
-		case KEYTYPE_EDITABLETEXT_CLEAR:
+	        case KEYTYPE_TEXTNAV:
+	        case KEYTYPE_EDIT:
+	        case KEYTYPE_SELECT:
+	        case KEYTYPE_TOGGLESELECT:
 			gok_keyboard_branch_editableTextAction 
 				(gok_main_get_current_keyboard(), pKey);
 			break;
@@ -2871,7 +2949,7 @@ gboolean gok_keyboard_validate_dynamic_keys (Accessible* pAccessibleForeground)
 		{
 			bHasGUIBranch = TRUE;
 		}
-		else if (pKey->Type == KEYTYPE_BRANCHEDITTEXT)
+		else if (pKey->Type == KEYTYPE_BRANCHCOMPOSE)
 		{
 			bHasEditTextBranch = TRUE;
 		}
@@ -2980,9 +3058,9 @@ gboolean gok_keyboard_validate_dynamic_keys (Accessible* pAccessibleForeground)
 							     "StyleTextDisabled");
 			}
 		}
-		else if (pKey->Type == KEYTYPE_BRANCHEDITTEXT)
+		else if (pKey->Type == KEYTYPE_BRANCHCOMPOSE)
 		{
-			if (gok_spy_get_accessibleWithEditableText() != NULL)
+			if (gok_spy_get_accessibleWithText() != NULL)
 			{
 				/* set this flag if the button state changes */
 				if (strcmp (gtk_widget_get_name(pKey->pButton), "StyleButtonBranchGUI") != 0)
@@ -3235,5 +3313,13 @@ gok_keyboard_parse_direction (const gchar *string)
 		return GOK_DIRECTION_SE;
 	else if (!strcmp (string, "fillwidth"))
 		return GOK_DIRECTION_FILL_EW;
+	else if (!strcmp (string, "narrower"))
+		return GOK_RESIZE_NARROWER;
+	else if (!strcmp (string, "wider"))
+		return GOK_RESIZE_WIDER;
+	else if (!strcmp (string, "shorter"))
+		return GOK_RESIZE_SHORTER;
+	else if (!strcmp (string, "taller"))
+		return GOK_RESIZE_TALLER;
 	return GOK_DIRECTION_NONE;
 }
