@@ -674,6 +674,46 @@ char WordCompleteAddNewWord (gchar* pWord)
 	return WordCompleteAddWord (pWord, PRIORITY_NEWWORD, STATE_TEMPORARY);
 }
 
+static Node*
+WordCompleteMatchingLeaf (Node *node, const gchar *pWord)
+{
+	Node *child = node->pFirstChildNode;
+	while (child != NULL) {
+		if (pWord && *pWord && (*pWord == child->letter)) {
+			if (child->priority > 0) 
+				return child;
+			else {
+				return WordCompleteMatchingLeaf (child, ++pWord);
+			}
+		}
+		else {
+			child = child->pNextNode;
+		}
+	}
+	return NULL;
+}
+
+/*
+ * WordCompleteIncrementFrequency
+ *
+ */
+gboolean 
+WordCompleteIncrementFrequency (const gchar *pWord)
+{
+	Node *node;
+	if (pWord) {
+		node = WordCompleteMatchingLeaf (&m_FirstNode, pWord);
+		fprintf (stderr, "node for %s: %p\n", pWord, node);
+		if (node && node->priority) {
+			++node->priority;
+			fprintf (stderr, "priority of %s is now %d\n", pWord, node->priority);
+			bWordCompleteModified = TRUE;
+			return TRUE;
+		}
+	}
+	return FALSE;
+}
+
 /*
  * WordCompleteSaveDictionary
  *
@@ -700,6 +740,7 @@ char WordCompleteSaveDictionary()
 	  return FALSE;
 	}
 
+	fprintf (stderr, "saving dictionary...\n");
 	/* iterate through each child node */
 	pNode = m_FirstNode.pFirstChildNode;
 	while (pNode != NULL)
